@@ -1,4 +1,4 @@
-// Copyright 2022-2023 The NATS Authors
+// Copyright 2022-2024 The NATS Authors
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
 // You may obtain a copy of the License at
@@ -28,108 +28,248 @@ import (
 
 type (
 
-	// JetStream contains CRUD methods to operate on a stream
-	// Create, update and get operations return 'Stream' interface,
-	// allowing operations on consumers
+	// JetStream is the top-level interface for interacting with JetStream.
+	// The capabilities of JetStream include:
 	//
-	// CreateOrUpdateConsumer, Consumer and DeleteConsumer are helper methods used to create/fetch/remove consumer without fetching stream (bypassing stream API)
+	// - Publishing messages to a stream using [Publisher].
+	// - Managing streams using [StreamManager].
+	// - Managing consumers using [StreamConsumerManager]. Those are the same
+	//   methods as on [Stream], but are available as a shortcut to a consumer
+	//   bypassing stream lookup.
+	// - Managing KeyValue stores using [KeyValueManager].
+	// - Managing Object Stores using [ObjectStoreManager].
 	//
-	// Client returns a JetStremClient, used to publish messages on a stream or fetch messages by sequence number
+	// JetStream can be created using [New], [NewWithAPIPrefix] or
+	// [NewWithDomain] methods.
 	JetStream interface {
-		// Returns *AccountInfo, containing details about the account associated with this JetStream connection
+		// AccountInfo fetches account information from the server, containing details
+		// about the account associated with this JetStream connection. If account is
+		// not enabled for JetStream, ErrJetStreamNotEnabledForAccount is returned. If
+		// the server does not have JetStream enabled, ErrJetStreamNotEnabled is
+		// returned.
 		AccountInfo(ctx context.Context) (*AccountInfo, error)
 
 		StreamConsumerManager
 		StreamManager
 		Publisher
 		KeyValueManager
+		ObjectStoreManager
 	}
 
+	// Publisher provides methods for publishing messages to a stream.
+	// It is available as a part of [JetStream] interface.
+	// The behavior of Publisher can be customized using [PublishOpt] options.
 	Publisher interface {
-		// Publish performs a synchronous publish to a stream and waits for ack from server
-		// It accepts subject name (which must be bound to a stream) and message data
+		// Publish performs a synchronous publish to a stream and waits for ack
+		// from server. It accepts subject name (which must be bound to a stream)
+		// and message payload.
 		Publish(ctx context.Context, subject string, payload []byte, opts ...PublishOpt) (*PubAck, error)
-		// PublishMsg performs a synchronous publish to a stream and waits for ack from server
-		// It accepts subject name (which must be bound to a stream) and nats.Message
+
+		// PublishMsg performs a synchronous publish to a stream and waits for
+		// ack from server. It accepts subject name (which must be bound to a
+		// stream) and nats.Message.
 		PublishMsg(ctx context.Context, msg *nats.Msg, opts ...PublishOpt) (*PubAck, error)
-		// PublishAsync performs an asynchronous publish to a stream and returns [PubAckFuture] interface
-		// It accepts subject name (which must be bound to a stream) and message data
+
+		// PublishAsync performs a publish to a stream and returns
+		// [PubAckFuture] interface, not blocking while waiting for an
+		// acknowledgement. It accepts subject name (which must be bound to a
+		// stream) and message payload.
+		//
+		// PublishAsync does not guarantee that the message has been
+		// received by the server. It only guarantees that the message has been
+		// sent to the server and thus messages can be stored in the stream
+		// out of order in case of retries.
 		PublishAsync(subject string, payload []byte, opts ...PublishOpt) (PubAckFuture, error)
-		// PublishMsgAsync performs an asynchronous publish to a stream and returns [PubAckFuture] interface
-		// It accepts subject name (which must be bound to a stream) and nats.Message
+
+		// PublishMsgAsync performs a publish to a stream and returns
+		// [PubAckFuture] interface, not blocking while waiting for an
+		// acknowledgement. It accepts subject name (which must
+		// be bound to a stream) and nats.Message.
+		//
+		// PublishMsgAsync does not guarantee that the message has been
+		// sent to the server and thus messages can be stored in the stream
+		// received by the server. It only guarantees that the message has been
+		// out of order in case of retries.
 		PublishMsgAsync(msg *nats.Msg, opts ...PublishOpt) (PubAckFuture, error)
-		// PublishAsyncPending returns the number of async publishes outstanding for this context
+
+		// PublishAsyncPending returns the number of async publishes outstanding
+		// for this context. An outstanding publish is one that has been
+		// sent by the publisher but has not yet received an ack.
 		PublishAsyncPending() int
-		// PublishAsyncComplete returns a channel that will be closed when all outstanding messages are ack'd
+
+		// PublishAsyncComplete returns a channel that will be closed when all
+		// outstanding asynchronously published messages are acknowledged by the
+		// server.
 		PublishAsyncComplete() <-chan struct{}
+
+		// CleanupPublisher will cleanup the publishing side of JetStreamContext.
+		//
+		// This will unsubscribe from the internal reply subject if needed.
+		// All pending async publishes will fail with ErrJetStreamContextClosed.
+		//
+		// If an error handler was provided, it will be called for each pending async
+		// publish and PublishAsyncComplete will be closed.
+		//
+		// After completing JetStreamContext is still usable - internal subscription
+		// will be recreated on next publish, but the acks from previous publishes will
+		// be lost.
+		CleanupPublisher()
 	}
 
+	// StreamManager provides CRUD API for managing streams. It is available as
+	// a part of [JetStream] interface. CreateStream, UpdateStream,
+	// CreateOrUpdateStream and Stream methods return a [Stream] interface, allowing
+	// to operate on a stream.
 	StreamManager interface {
-		// CreateStream creates a new stream with given config and returns a hook to operate on it
+		// CreateStream creates a new stream with given config and returns an
+		// interface to operate on it. If stream with given name already exists
+		// and its configuration differs from the provided one,
+		// ErrStreamNameAlreadyInUse is returned.
 		CreateStream(ctx context.Context, cfg StreamConfig) (Stream, error)
-		// UpdateStream updates an existing stream
+
+		// UpdateStream updates an existing stream. If stream does not exist,
+		// ErrStreamNotFound is returned.
 		UpdateStream(ctx context.Context, cfg StreamConfig) (Stream, error)
-		// Stream returns a [Stream] hook for a given stream name
+
+		// CreateOrUpdateStream creates a stream with given config. If stream
+		// already exists, it will be updated (if possible).
+		CreateOrUpdateStream(ctx context.Context, cfg StreamConfig) (Stream, error)
+
+		// Stream fetches [StreamInfo] and returns a [Stream] interface for a given stream name.
+		// If stream does not exist, ErrStreamNotFound is returned.
 		Stream(ctx context.Context, stream string) (Stream, error)
-		// StreamNameBySubject returns a stream name stream listening on given subject
+
+		// StreamNameBySubject returns a stream name stream listening on given
+		// subject. If no stream is bound to given subject, ErrStreamNotFound
+		// is returned.
 		StreamNameBySubject(ctx context.Context, subject string) (string, error)
-		// DeleteStream removes a stream with given name
+
+		// DeleteStream removes a stream with given name. If stream does not
+		// exist, ErrStreamNotFound is returned.
 		DeleteStream(ctx context.Context, stream string) error
-		// ListStreams returns StreamInfoLister enabling iterating over a channel of stream infos
+
+		// ListStreams returns StreamInfoLister, enabling iterating over a
+		// channel of stream infos.
 		ListStreams(context.Context, ...StreamListOpt) StreamInfoLister
-		// StreamNames returns a  StreamNameLister enabling iterating over a channel of stream names
+
+		// StreamNames returns a  StreamNameLister, enabling iterating over a
+		// channel of stream names.
 		StreamNames(context.Context, ...StreamListOpt) StreamNameLister
 	}
 
+	// StreamConsumerManager provides CRUD API for managing consumers. It is
+	// available as a part of [JetStream] interface. This is an alternative to
+	// [Stream] interface, allowing to bypass stream lookup. CreateConsumer,
+	// UpdateConsumer, CreateOrUpdateConsumer and Consumer methods return a
+	// [Consumer] interface, allowing to operate on a consumer (e.g. consume
+	// messages).
 	StreamConsumerManager interface {
-		// CreateOrUpdateConsumer creates a consumer on a given stream with given config.
-		// If consumer already exists, it will be updated (if possible).
-		// Consumer interface is returned, serving as a hook to operate on a consumer (e.g. fetch messages)
+		// CreateOrUpdateConsumer creates a consumer on a given stream with
+		// given config. If consumer already exists, it will be updated (if
+		// possible). Consumer interface is returned, allowing to operate on a
+		// consumer (e.g. fetch messages).
 		CreateOrUpdateConsumer(ctx context.Context, stream string, cfg ConsumerConfig) (Consumer, error)
-		// CreateConsumer creates a consumer on a given stream with given config.
-		// If consumer already exists, ErrConsumerExists is returned.
-		// Consumer interface is returned, serving as a hook to operate on a consumer (e.g. fetch messages)
+
+		// CreateConsumer creates a consumer on a given stream with given
+		// config. If consumer already exists and the provided configuration
+		// differs from its configuration, ErrConsumerExists is returned. If the
+		// provided configuration is the same as the existing consumer, the
+		// existing consumer is returned. Consumer interface is returned,
+		// allowing to operate on a consumer (e.g. fetch messages).
 		CreateConsumer(ctx context.Context, stream string, cfg ConsumerConfig) (Consumer, error)
-		// UpdateConsumer updates an existing consumer.
-		// If consumer does not exist, ErrConsumerDoesNotExist is returned.
-		// Consumer interface is returned, serving as a hook to operate on a consumer (e.g. fetch messages)
+
+		// UpdateConsumer updates an existing consumer. If consumer does not
+		// exist, ErrConsumerDoesNotExist is returned. Consumer interface is
+		// returned, allowing to operate on a consumer (e.g. fetch messages).
 		UpdateConsumer(ctx context.Context, stream string, cfg ConsumerConfig) (Consumer, error)
-		// OrderedConsumer returns an OrderedConsumer instance.
-		// OrderedConsumer allows fetching messages from a stream (just like standard consumer),
-		// for in order delivery of messages. Underlying consumer is re-created when necessary,
-		// without additional client code.
+
+		// OrderedConsumer returns an OrderedConsumer instance. OrderedConsumer
+		// are managed by the library and provide a simple way to consume
+		// messages from a stream. Ordered consumers are ephemeral in-memory
+		// pull consumers and are resilient to deletes and restarts.
 		OrderedConsumer(ctx context.Context, stream string, cfg OrderedConsumerConfig) (Consumer, error)
-		// Consumer returns a hook to an existing consumer, allowing processing of messages
+
+		// Consumer returns an interface to an existing consumer, allowing processing
+		// of messages. If consumer does not exist, ErrConsumerNotFound is
+		// returned.
 		Consumer(ctx context.Context, stream string, consumer string) (Consumer, error)
-		// DeleteConsumer removes a consumer with given name from a stream
+
+		// DeleteConsumer removes a consumer with given name from a stream.
+		// If consumer does not exist, ErrConsumerNotFound is returned.
 		DeleteConsumer(ctx context.Context, stream string, consumer string) error
 	}
 
+	// StreamListOpt is a functional option for [StreamManager.ListStreams] and
+	// [StreamManager.StreamNames] methods.
 	StreamListOpt func(*streamsRequest) error
 
-	// AccountInfo contains info about the JetStream usage from the current account.
+	// AccountInfo contains information about the JetStream usage from the
+	// current account.
 	AccountInfo struct {
-		Memory    uint64        `json:"memory"`
-		Store     uint64        `json:"storage"`
-		Streams   int           `json:"streams"`
-		Consumers int           `json:"consumers"`
-		Domain    string        `json:"domain"`
-		API       APIStats      `json:"api"`
-		Limits    AccountLimits `json:"limits"`
+		// Tier is the current account usage tier.
+		Tier
+
+		// Domain is the domain name associated with this account.
+		Domain string `json:"domain"`
+
+		// API is the API usage statistics for this account.
+		API APIStats `json:"api"`
+
+		// Tiers is the list of available tiers for this account.
+		Tiers map[string]Tier `json:"tiers"`
+	}
+
+	// Tier represents a JetStream account usage tier.
+	Tier struct {
+		// Memory is the memory storage being used for Stream Message storage.
+		Memory uint64 `json:"memory"`
+
+		// Store is the disk storage being used for Stream Message storage.
+		Store uint64 `json:"storage"`
+
+		// ReservedMemory is the number of bytes reserved for memory usage by
+		// this account on the server
+		ReservedMemory uint64 `json:"reserved_memory"`
+
+		// ReservedStore is the number of bytes reserved for disk usage by this
+		// account on the server
+		ReservedStore uint64 `json:"reserved_storage"`
+
+		// Streams is the number of streams currently defined for this account.
+		Streams int `json:"streams"`
+
+		// Consumers is the number of consumers currently defined for this
+		// account.
+		Consumers int `json:"consumers"`
+
+		// Limits are the JetStream limits for this account.
+		Limits AccountLimits `json:"limits"`
 	}
 
 	// APIStats reports on API calls to JetStream for this account.
 	APIStats struct {
-		Total  uint64 `json:"total"`
+		// Total is the total number of API calls.
+		Total uint64 `json:"total"`
+
+		// Errors is the total number of API errors.
 		Errors uint64 `json:"errors"`
 	}
 
 	// AccountLimits includes the JetStream limits of the current account.
 	AccountLimits struct {
-		MaxMemory    int64 `json:"max_memory"`
-		MaxStore     int64 `json:"max_storage"`
-		MaxStreams   int   `json:"max_streams"`
-		MaxConsumers int   `json:"max_consumers"`
+		// MaxMemory is the maximum amount of memory available for this account.
+		MaxMemory int64 `json:"max_memory"`
+
+		// MaxStore is the maximum amount of disk storage available for this
+		// account.
+		MaxStore int64 `json:"max_storage"`
+
+		// MaxStreams is the maximum number of streams allowed for this account.
+		MaxStreams int `json:"max_streams"`
+
+		// MaxConsumers is the maximum number of consumers allowed for this
+		// account.
+		MaxConsumers int `json:"max_consumers"`
 	}
 
 	jetStream struct {
@@ -139,21 +279,30 @@ type (
 		publisher *jetStreamClient
 	}
 
+	// JetStreamOpt is a functional option for [New], [NewWithAPIPrefix] and
+	// [NewWithDomain] methods.
 	JetStreamOpt func(*jsOpts) error
 
 	jsOpts struct {
-		publisherOpts asyncPublisherOpts
-		apiPrefix     string
-		clientTrace   *ClientTrace
+		publisherOpts  asyncPublisherOpts
+		apiPrefix      string
+		replyPrefix    string
+		replyPrefixLen int
+		clientTrace    *ClientTrace
 	}
 
-	// ClientTrace can be used to trace API interactions for the JetStream Context.
+	// ClientTrace can be used to trace API interactions for [JetStream].
 	ClientTrace struct {
-		RequestSent      func(subj string, payload []byte)
+		// RequestSent is called when an API request is sent to the server.
+		RequestSent func(subj string, payload []byte)
+
+		// ResponseReceived is called when a response is received from the
+		// server.
 		ResponseReceived func(subj string, payload []byte, hdr nats.Header)
 	}
 	streamInfoResponse struct {
 		apiResponse
+		apiPaged
 		*StreamInfo
 	}
 
@@ -167,11 +316,17 @@ type (
 		Success bool `json:"success,omitempty"`
 	}
 
+	// StreamInfoLister is used to iterate over a channel of stream infos.
+	// Err method can be used to check for errors encountered during iteration.
+	// Info channel is always closed and therefore can be used in a range loop.
 	StreamInfoLister interface {
 		Info() <-chan *StreamInfo
 		Err() error
 	}
 
+	// StreamNameLister is used to iterate over a channel of stream names.
+	// Err method can be used to check for errors encountered during iteration.
+	// Name channel is always closed and therefore can be used in a range loop.
 	StreamNameLister interface {
 		Name() <-chan string
 		Err() error
@@ -180,6 +335,7 @@ type (
 	apiPagedRequest struct {
 		Offset int `json:"offset"`
 	}
+
 	streamLister struct {
 		js       *jetStream
 		offset   int
@@ -214,12 +370,14 @@ const defaultAPITimeout = 5 * time.Second
 var subjectRegexp = regexp.MustCompile(`^[^ >]*[>]?$`)
 
 // New returns a new JetStream instance.
+// It uses default API prefix ($JS.API) for JetStream API requests.
+// If a custom API prefix is required, use [NewWithAPIPrefix] or [NewWithDomain].
 //
 // Available options:
-// [WithClientTrace] - enables request/response tracing.
-// [WithPublishAsyncErrHandler] - sets error handler for async message publish.
-// [WithPublishAsyncMaxPending] - sets the maximum outstanding async publishes that can be inflight at one time.
-// [WithDirectGet] - specifies whether client should use direct get requests.
+//   - [WithClientTrace] - enables request/response tracing.
+//   - [WithPublishAsyncErrHandler] - sets error handler for async message publish.
+//   - [WithPublishAsyncMaxPending] - sets the maximum outstanding async publishes
+//     that can be inflight at one time.
 func New(nc *nats.Conn, opts ...JetStreamOpt) (JetStream, error) {
 	jsOpts := jsOpts{
 		apiPrefix: DefaultAPIPrefix,
@@ -227,6 +385,7 @@ func New(nc *nats.Conn, opts ...JetStreamOpt) (JetStream, error) {
 			maxpa: defaultAsyncPubAckInflight,
 		},
 	}
+	setReplyPrefix(nc, &jsOpts)
 	for _, opt := range opts {
 		if err := opt(&jsOpts); err != nil {
 			return nil, err
@@ -246,26 +405,38 @@ const (
 	defaultAsyncPubAckInflight = 4000
 )
 
-// NewWithAPIPrefix returns a new JetStream instance and sets the API prefix to be used in requests to JetStream API
+func setReplyPrefix(nc *nats.Conn, jsOpts *jsOpts) {
+	jsOpts.replyPrefix = nats.InboxPrefix
+	if nc.Opts.InboxPrefix != "" {
+		jsOpts.replyPrefix = nc.Opts.InboxPrefix + "."
+	}
+	// Add 1 for the dot separator.
+	jsOpts.replyPrefixLen = len(jsOpts.replyPrefix) + aReplyTokensize + 1
+
+}
+
+// NewWithAPIPrefix returns a new JetStream instance and sets the API prefix to be used in requests to JetStream API.
+// The API prefix will be used in API requests to JetStream, e.g. <prefix>.STREAM.INFO.<stream>.
 //
 // Available options:
-// [WithClientTrace] - enables request/response tracing
-// [WithPublishAsyncErrHandler] - sets error handler for async message publish
-// [WithPublishAsyncMaxPending] - sets the maximum outstanding async publishes that can be inflight at one time.
-// [WithDirectGet] - specifies whether client should use direct get requests.
+//   - [WithClientTrace] - enables request/response tracing.
+//   - [WithPublishAsyncErrHandler] - sets error handler for async message publish.
+//   - [WithPublishAsyncMaxPending] - sets the maximum outstanding async publishes
+//     that can be inflight at one time.
 func NewWithAPIPrefix(nc *nats.Conn, apiPrefix string, opts ...JetStreamOpt) (JetStream, error) {
 	jsOpts := jsOpts{
 		publisherOpts: asyncPublisherOpts{
 			maxpa: defaultAsyncPubAckInflight,
 		},
 	}
+	setReplyPrefix(nc, &jsOpts)
 	for _, opt := range opts {
 		if err := opt(&jsOpts); err != nil {
 			return nil, err
 		}
 	}
 	if apiPrefix == "" {
-		return nil, fmt.Errorf("API prefix cannot be empty")
+		return nil, errors.New("API prefix cannot be empty")
 	}
 	if !strings.HasSuffix(apiPrefix, ".") {
 		jsOpts.apiPrefix = fmt.Sprintf("%s.", apiPrefix)
@@ -278,26 +449,28 @@ func NewWithAPIPrefix(nc *nats.Conn, apiPrefix string, opts ...JetStreamOpt) (Je
 	return js, nil
 }
 
-// NewWithDomain returns a new JetStream instance and sets the domain name token used when sending JetStream requests
+// NewWithDomain returns a new JetStream instance and sets the domain name token used when sending JetStream requests.
+// The domain name token will be used in API requests to JetStream, e.g. $JS.<domain>.API.STREAM.INFO.<stream>.
 //
 // Available options:
-// [WithClientTrace] - enables request/response tracing
-// [WithPublishAsyncErrHandler] - sets error handler for async message publish
-// [WithPublishAsyncMaxPending] - sets the maximum outstanding async publishes that can be inflight at one time.
-// [WithDirectGet] - specifies whether client should use direct get requests.
+//   - [WithClientTrace] - enables request/response tracing.
+//   - [WithPublishAsyncErrHandler] - sets error handler for async message publish.
+//   - [WithPublishAsyncMaxPending] - sets the maximum outstanding async publishes
+//     that can be inflight at one time.
 func NewWithDomain(nc *nats.Conn, domain string, opts ...JetStreamOpt) (JetStream, error) {
 	jsOpts := jsOpts{
 		publisherOpts: asyncPublisherOpts{
 			maxpa: defaultAsyncPubAckInflight,
 		},
 	}
+	setReplyPrefix(nc, &jsOpts)
 	for _, opt := range opts {
 		if err := opt(&jsOpts); err != nil {
 			return nil, err
 		}
 	}
 	if domain == "" {
-		return nil, fmt.Errorf("domain cannot be empty")
+		return nil, errors.New("domain cannot be empty")
 	}
 	jsOpts.apiPrefix = fmt.Sprintf(jsDomainT, domain)
 	js := &jetStream{
@@ -308,7 +481,9 @@ func NewWithDomain(nc *nats.Conn, domain string, opts ...JetStreamOpt) (JetStrea
 	return js, nil
 }
 
-// CreateStream creates a new stream with given config and returns a hook to operate on it
+// CreateStream creates a new stream with given config and returns an
+// interface to operate on it. If stream with given name already exists,
+// ErrStreamNameAlreadyInUse is returned.
 func (js *jetStream) CreateStream(ctx context.Context, cfg StreamConfig) (Stream, error) {
 	if err := validateStreamName(cfg.Name); err != nil {
 		return nil, err
@@ -409,7 +584,8 @@ func (ss *StreamSource) copy() *StreamSource {
 	return &nss
 }
 
-// UpdateStream updates an existing stream
+// UpdateStream updates an existing stream. If stream does not exist,
+// ErrStreamNotFound is returned.
 func (js *jetStream) UpdateStream(ctx context.Context, cfg StreamConfig) (Stream, error) {
 	if err := validateStreamName(cfg.Name); err != nil {
 		return nil, err
@@ -460,7 +636,22 @@ func (js *jetStream) UpdateStream(ctx context.Context, cfg StreamConfig) (Stream
 	}, nil
 }
 
-// Stream returns a [Stream] hook for a given stream name
+// CreateOrUpdateStream creates a stream with given config. If stream
+// already exists, it will be updated (if possible).
+func (js *jetStream) CreateOrUpdateStream(ctx context.Context, cfg StreamConfig) (Stream, error) {
+	s, err := js.UpdateStream(ctx, cfg)
+	if err != nil {
+		if !errors.Is(err, ErrStreamNotFound) {
+			return nil, err
+		}
+		return js.CreateStream(ctx, cfg)
+	}
+
+	return s, nil
+}
+
+// Stream fetches [StreamInfo] and returns a [Stream] interface for a given stream name.
+// If stream does not exist, ErrStreamNotFound is returned.
 func (js *jetStream) Stream(ctx context.Context, name string) (Stream, error) {
 	if err := validateStreamName(name); err != nil {
 		return nil, err
@@ -513,9 +704,10 @@ func (js *jetStream) DeleteStream(ctx context.Context, name string) error {
 	return nil
 }
 
-// CreateOrUpdateConsumer creates a consumer on a given stream with given config
-// This operation is idempotent - if a consumer already exists, it will be a no-op (or error if configs do not match)
-// Consumer interface is returned, serving as a hook to operate on a consumer (e.g. fetch messages)
+// CreateOrUpdateConsumer creates a consumer on a given stream with
+// given config. If consumer already exists, it will be updated (if
+// possible). Consumer interface is returned, allowing to operate on a
+// consumer (e.g. fetch messages).
 func (js *jetStream) CreateOrUpdateConsumer(ctx context.Context, stream string, cfg ConsumerConfig) (Consumer, error) {
 	if err := validateStreamName(stream); err != nil {
 		return nil, err
@@ -523,6 +715,12 @@ func (js *jetStream) CreateOrUpdateConsumer(ctx context.Context, stream string,
 	return upsertConsumer(ctx, js, stream, cfg, consumerActionCreateOrUpdate)
 }
 
+// CreateConsumer creates a consumer on a given stream with given
+// config. If consumer already exists and the provided configuration
+// differs from its configuration, ErrConsumerExists is returned. If the
+// provided configuration is the same as the existing consumer, the
+// existing consumer is returned. Consumer interface is returned,
+// allowing to operate on a consumer (e.g. fetch messages).
 func (js *jetStream) CreateConsumer(ctx context.Context, stream string, cfg ConsumerConfig) (Consumer, error) {
 	if err := validateStreamName(stream); err != nil {
 		return nil, err
@@ -530,6 +728,9 @@ func (js *jetStream) CreateConsumer(ctx context.Context, stream string, cfg Cons
 	return upsertConsumer(ctx, js, stream, cfg, consumerActionCreate)
 }
 
+// UpdateConsumer updates an existing consumer. If consumer does not
+// exist, ErrConsumerDoesNotExist is returned. Consumer interface is
+// returned, allowing to operate on a consumer (e.g. fetch messages).
 func (js *jetStream) UpdateConsumer(ctx context.Context, stream string, cfg ConsumerConfig) (Consumer, error) {
 	if err := validateStreamName(stream); err != nil {
 		return nil, err
@@ -537,6 +738,10 @@ func (js *jetStream) UpdateConsumer(ctx context.Context, stream string, cfg Cons
 	return upsertConsumer(ctx, js, stream, cfg, consumerActionUpdate)
 }
 
+// OrderedConsumer returns an OrderedConsumer instance. OrderedConsumer
+// are managed by the library and provide a simple way to consume
+// messages from a stream. Ordered consumers are ephemeral in-memory
+// pull consumers and are resilient to deletes and restarts.
 func (js *jetStream) OrderedConsumer(ctx context.Context, stream string, cfg OrderedConsumerConfig) (Consumer, error) {
 	if err := validateStreamName(stream); err != nil {
 		return nil, err
@@ -548,18 +753,19 @@ func (js *jetStream) OrderedConsumer(ctx context.Context, stream string, cfg Ord
 		namePrefix: nuid.Next(),
 		doReset:    make(chan struct{}, 1),
 	}
-	if cfg.OptStartSeq != 0 {
-		oc.cursor.streamSeq = cfg.OptStartSeq - 1
-	}
-	err := oc.reset()
+	consCfg := oc.getConsumerConfig()
+	cons, err := js.CreateOrUpdateConsumer(ctx, stream, *consCfg)
 	if err != nil {
 		return nil, err
 	}
+	oc.currentConsumer = cons.(*pullConsumer)
 
 	return oc, nil
 }
 
-// Consumer returns a hook to an existing consumer, allowing processing of messages
+// Consumer returns an interface to an existing consumer, allowing processing
+// of messages. If consumer does not exist, ErrConsumerNotFound is
+// returned.
 func (js *jetStream) Consumer(ctx context.Context, stream string, name string) (Consumer, error) {
 	if err := validateStreamName(stream); err != nil {
 		return nil, err
@@ -567,7 +773,8 @@ func (js *jetStream) Consumer(ctx context.Context, stream string, name string) (
 	return getConsumer(ctx, js, stream, name)
 }
 
-// DeleteConsumer removes a consumer with given name from a stream
+// DeleteConsumer removes a consumer with given name from a stream.
+// If consumer does not exist, ErrConsumerNotFound is returned.
 func (js *jetStream) DeleteConsumer(ctx context.Context, stream string, name string) error {
 	if err := validateStreamName(stream); err != nil {
 		return err
@@ -579,7 +786,7 @@ func validateStreamName(stream string) error {
 	if stream == "" {
 		return ErrStreamNameRequired
 	}
-	if strings.Contains(stream, ".") {
+	if strings.ContainsAny(stream, ">*. /\\") {
 		return fmt.Errorf("%w: '%s'", ErrInvalidStreamName, stream)
 	}
 	return nil
@@ -589,12 +796,19 @@ func validateSubject(subject string) error {
 	if subject == "" {
 		return fmt.Errorf("%w: %s", ErrInvalidSubject, "subject cannot be empty")
 	}
-	if !subjectRegexp.MatchString(subject) {
+	if subject[0] == '.' || subject[len(subject)-1] == '.' || !subjectRegexp.MatchString(subject) {
 		return fmt.Errorf("%w: %s", ErrInvalidSubject, subject)
 	}
 	return nil
 }
 
+// AccountInfo fetches account information from the server, containing details
+// about the account associated with this JetStream connection. If account is
+// not enabled for JetStream, ErrJetStreamNotEnabledForAccount is returned.
+//
+// If the server does not have JetStream enabled, ErrJetStreamNotEnabled is
+// returned (for a single server setup). For clustered topologies, AccountInfo
+// will time out.
 func (js *jetStream) AccountInfo(ctx context.Context) (*AccountInfo, error) {
 	ctx, cancel := wrapContextWithoutDeadline(ctx)
 	if cancel != nil {
@@ -622,10 +836,8 @@ func (js *jetStream) AccountInfo(ctx context.Context) (*AccountInfo, error) {
 	return &resp.AccountInfo, nil
 }
 
-// ListStreams returns StreamInfoLister enabling iterating over a channel of stream infos
-//
-// Available options:
-// [WithStreamListSubject] - allows filtering returned streams by provided subject
+// ListStreams returns StreamInfoLister, enabling iterating over a
+// channel of stream infos.
 func (js *jetStream) ListStreams(ctx context.Context, opts ...StreamListOpt) StreamInfoLister {
 	l := &streamLister{
 		js:      js,
@@ -678,10 +890,8 @@ func (s *streamLister) Err() error {
 	return s.err
 }
 
-// StreamNames returns a [StreamNameLister] enabling iterating over a channel of stream names
-//
-// Available options:
-// [WithStreamListSubject] - allows filtering returned streams by provided subject
+// StreamNames returns a  StreamNameLister, enabling iterating over a
+// channel of stream names.
 func (js *jetStream) StreamNames(ctx context.Context, opts ...StreamListOpt) StreamNameLister {
 	l := &streamLister{
 		js:    js,
@@ -724,6 +934,9 @@ func (js *jetStream) StreamNames(ctx context.Context, opts ...StreamListOpt) Str
 	return l
 }
 
+// StreamNameBySubject returns a stream name stream listening on given
+// subject. If no stream is bound to given subject, ErrStreamNotFound
+// is returned.
 func (js *jetStream) StreamNameBySubject(ctx context.Context, subject string) (string, error) {
 	ctx, cancel := wrapContextWithoutDeadline(ctx)
 	if cancel != nil {
@@ -832,3 +1045,52 @@ func wrapContextWithoutDeadline(ctx context.Context) (context.Context, context.C
 	}
 	return context.WithTimeout(ctx, defaultAPITimeout)
 }
+
+// CleanupPublisher will cleanup the publishing side of JetStreamContext.
+//
+// This will unsubscribe from the internal reply subject if needed.
+// All pending async publishes will fail with ErrJetStreamContextClosed.
+//
+// If an error handler was provided, it will be called for each pending async
+// publish and PublishAsyncComplete will be closed.
+//
+// After completing JetStreamContext is still usable - internal subscription
+// will be recreated on next publish, but the acks from previous publishes will
+// be lost.
+func (js *jetStream) CleanupPublisher() {
+	js.cleanupReplySub()
+	js.publisher.Lock()
+	errCb := js.publisher.aecb
+	for id, paf := range js.publisher.acks {
+		paf.err = ErrJetStreamPublisherClosed
+		if paf.errCh != nil {
+			paf.errCh <- paf.err
+		}
+		if errCb != nil {
+			// call error handler after releasing the mutex to avoid contention
+			defer errCb(js, paf.msg, ErrJetStreamPublisherClosed)
+		}
+		delete(js.publisher.acks, id)
+	}
+	if js.publisher.doneCh != nil {
+		close(js.publisher.doneCh)
+		js.publisher.doneCh = nil
+	}
+	js.publisher.Unlock()
+}
+
+func (js *jetStream) cleanupReplySub() {
+	if js.publisher == nil {
+		return
+	}
+	js.publisher.Lock()
+	if js.publisher.replySub != nil {
+		js.publisher.replySub.Unsubscribe()
+		js.publisher.replySub = nil
+	}
+	if js.publisher.connStatusCh != nil {
+		close(js.publisher.connStatusCh)
+		js.publisher.connStatusCh = nil
+	}
+	js.publisher.Unlock()
+}
