@@ -1,4 +1,4 @@
-// Copyright 2022-2023 The NATS Authors
+// Copyright 2022-2024 The NATS Authors
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
 // You may obtain a copy of the License at
@@ -14,7 +14,6 @@
 package jetstream
 
 import (
-	"context"
 	"encoding/json"
 	"errors"
 	"fmt"
@@ -24,43 +23,70 @@ import (
 	"time"
 
 	"github.com/nats-io/nats.go"
+	"github.com/nats-io/nats.go/internal/syncx"
 	"github.com/nats-io/nuid"
 )
 
 type (
 	// MessagesContext supports iterating over a messages on a stream.
+	// It is returned by [Consumer.Messages] method.
 	MessagesContext interface {
-		// Next retreives next message on a stream. It will block until the next message is available.
+		// Next retrieves next message on a stream. It will block until the next
+		// message is available. If the context is canceled, Next will return
+		// ErrMsgIteratorClosed error.
 		Next() (Msg, error)
-		// Stop closes the iterator and cancels subscription.
+
+		// Stop unsubscribes from the stream and cancels subscription. Calling
+		// Next after calling Stop will return ErrMsgIteratorClosed error.
+		// All messages that are already in the buffer are discarded.
 		Stop()
+
+		// Drain unsubscribes from the stream and cancels subscription. All
+		// messages that are already in the buffer will be available on
+		// subsequent calls to Next. After the buffer is drained, Next will
+		// return ErrMsgIteratorClosed error.
+		Drain()
 	}
 
+	// ConsumeContext supports processing incoming messages from a stream.
+	// It is returned by [Consumer.Consume] method.
 	ConsumeContext interface {
+		// Stop unsubscribes from the stream and cancels subscription.
+		// No more messages will be received after calling this method.
+		// All messages that are already in the buffer are discarded.
 		Stop()
+
+		// Drain unsubscribes from the stream and cancels subscription.
+		// All messages that are already in the buffer will be processed in callback function.
+		Drain()
+
+		// Closed returns a channel that is closed when the consuming is
+		// fully stopped/drained. When the channel is closed, no more messages
+		// will be received and processing is complete.
+		Closed() <-chan struct{}
 	}
 
-	// MessageHandler is a handler function used as callback in [Consume]
+	// MessageHandler is a handler function used as callback in [Consume].
 	MessageHandler func(msg Msg)
 
-	// PullConsumeOpt represent additional options used in [Consume] for pull consumers
+	// PullConsumeOpt represent additional options used in [Consume] for pull consumers.
 	PullConsumeOpt interface {
 		configureConsume(*consumeOpts) error
 	}
 
-	// PullMessagesOpt represent additional options used in [Messages] for pull consumers
+	// PullMessagesOpt represent additional options used in [Messages] for pull consumers.
 	PullMessagesOpt interface {
 		configureMessages(*consumeOpts) error
 	}
 
 	pullConsumer struct {
 		sync.Mutex
-		jetStream     *jetStream
-		stream        string
-		durable       bool
-		name          string
-		info          *ConsumerInfo
-		subscriptions map[string]*pullSubscription
+		jetStream *jetStream
+		stream    string
+		durable   bool
+		name      string
+		info      *ConsumerInfo
+		subs      syncx.Map[string, *pullSubscription]
 	}
 
 	pullRequest struct {
@@ -82,6 +108,7 @@ type (
 		ThresholdBytes          int
 		StopAfter               int
 		stopAfterMsgsLeft       chan int
+		notifyOnReconnect       bool
 	}
 
 	ConsumeErrHandlerFunc func(consumeCtx ConsumeContext, err error)
@@ -95,13 +122,15 @@ type (
 		errs              chan error
 		pending           pendingMsgs
 		hbMonitor         *hbMonitor
-		fetchInProgress   uint32
-		closed            uint32
+		fetchInProgress   atomic.Uint32
+		closed            atomic.Uint32
+		draining          atomic.Uint32
 		done              chan struct{}
 		connStatusChanged chan nats.Status
 		fetchNext         chan *pullRequest
 		consumeOpts       *consumeOpts
 		delivered         int
+		closedCh          chan struct{}
 	}
 
 	pendingMsgs struct {
@@ -115,6 +144,7 @@ type (
 	}
 
 	fetchResult struct {
+		sync.Mutex
 		msgs chan Msg
 		err  error
 		done bool
@@ -132,7 +162,6 @@ type (
 const (
 	DefaultMaxMessages = 500
 	DefaultExpires     = 30 * time.Second
-	DefaultHeartbeat   = 5 * time.Second
 	unset              = -1
 )
 
@@ -143,21 +172,16 @@ func min(x, y int) int {
 	return y
 }
 
-// Consume returns a ConsumeContext, allowing for processing incoming messages from a stream in a given callback function.
+// Consume can be used to continuously receive messages and handle them
+// with the provided callback function. Consume cannot be used concurrently
+// when using ordered consumer.
 //
-// Available options:
-// [PullMaxMessages] - sets maximum number of messages stored in a buffer, default is set to 100
-// [PullMaxBytes] - sets maximum number of bytes stored in a buffer
-// [PullExpiry] - sets a timeout for individual batch request, default is set to 30 seconds
-// [PullHeartbeat] - sets an idle heartbeat setting for a pull request, default is set to 5s
-// [ConsumeErrHandler] - sets custom consume error callback handler
-// [PullThresholdMessages] - sets the message count on which Consume will trigger new pull request to the server
-// [PullThresholdBytes] - sets the byte count on which Consume will trigger new pull request to the server
+// See [Consumer.Consume] for more details.
 func (p *pullConsumer) Consume(handler MessageHandler, opts ...PullConsumeOpt) (ConsumeContext, error) {
 	if handler == nil {
 		return nil, ErrHandlerRequired
 	}
-	consumeOpts, err := parseConsumeOpts(opts...)
+	consumeOpts, err := parseConsumeOpts(false, opts...)
 	if err != nil {
 		return nil, fmt.Errorf("%w: %s", ErrInvalidOption, err)
 	}
@@ -165,12 +189,7 @@ func (p *pullConsumer) Consume(handler MessageHandler, opts ...PullConsumeOpt) (
 
 	subject := apiSubj(p.jetStream.apiPrefix, fmt.Sprintf(apiRequestNextT, p.stream, p.name))
 
-	// for single consume, use empty string as id
-	// this is useful for ordered consumer, where only a single subscription is valid
-	var consumeID string
-	if len(p.subscriptions) > 0 {
-		consumeID = nuid.Next()
-	}
+	consumeID := nuid.Next()
 	sub := &pullSubscription{
 		id:          consumeID,
 		consumer:    p,
@@ -183,7 +202,7 @@ func (p *pullConsumer) Consume(handler MessageHandler, opts ...PullConsumeOpt) (
 
 	sub.hbMonitor = sub.scheduleHeartbeatCheck(consumeOpts.Heartbeat)
 
-	p.subscriptions[sub.id] = sub
+	p.subs.Store(sub.id, sub)
 	p.Unlock()
 
 	internalHandler := func(msg *nats.Msg) {
@@ -192,12 +211,17 @@ func (p *pullConsumer) Consume(handler MessageHandler, opts ...PullConsumeOpt) (
 		}
 		userMsg, msgErr := checkMsg(msg)
 		if !userMsg && msgErr == nil {
+			if sub.hbMonitor != nil {
+				sub.hbMonitor.Reset(2 * consumeOpts.Heartbeat)
+			}
 			return
 		}
 		defer func() {
 			sub.Lock()
 			sub.checkPending()
-			sub.hbMonitor.Reset(2 * consumeOpts.Heartbeat)
+			if sub.hbMonitor != nil {
+				sub.hbMonitor.Reset(2 * consumeOpts.Heartbeat)
+			}
 			sub.Unlock()
 		}()
 		if !userMsg {
@@ -211,7 +235,7 @@ func (p *pullConsumer) Consume(handler MessageHandler, opts ...PullConsumeOpt) (
 			sub.Unlock()
 
 			if err != nil {
-				if atomic.LoadUint32(&sub.closed) == 1 {
+				if sub.closed.Load() == 1 {
 					return
 				}
 				if sub.consumeOpts.ErrHandler != nil {
@@ -236,6 +260,18 @@ func (p *pullConsumer) Consume(handler MessageHandler, opts ...PullConsumeOpt) (
 	if err != nil {
 		return nil, err
 	}
+	sub.subscription.SetClosedHandler(func(sid string) func(string) {
+		return func(subject string) {
+			p.subs.Delete(sid)
+			sub.draining.CompareAndSwap(1, 0)
+			sub.Lock()
+			if sub.closedCh != nil {
+				close(sub.closedCh)
+				sub.closedCh = nil
+			}
+			sub.Unlock()
+		}
+	}(sub.id))
 
 	sub.Lock()
 	// initial pull
@@ -257,7 +293,7 @@ func (p *pullConsumer) Consume(handler MessageHandler, opts ...PullConsumeOpt) (
 	go func() {
 		isConnected := true
 		for {
-			if atomic.LoadUint32(&sub.closed) == 1 {
+			if sub.closed.Load() == 1 {
 				return
 			}
 			select {
@@ -275,36 +311,19 @@ func (p *pullConsumer) Consume(handler MessageHandler, opts ...PullConsumeOpt) (
 					sub.Lock()
 					if !isConnected {
 						isConnected = true
-						// try fetching consumer info several times to make sure consumer is available after reconnect
-						for i := 0; i < 5; i++ {
-							ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
-							_, err := p.Info(ctx)
-							cancel()
-							if err == nil {
-								break
-							}
-							if err != nil {
-								if i == 4 {
-									sub.cleanup()
-									if sub.consumeOpts.ErrHandler != nil {
-										sub.consumeOpts.ErrHandler(sub, err)
-									}
-									sub.Unlock()
-									return
-								}
-							}
-							time.Sleep(5 * time.Second)
-						}
-						batchSize := sub.consumeOpts.MaxMessages
-						if sub.consumeOpts.StopAfter > 0 {
-							batchSize = min(batchSize, sub.consumeOpts.StopAfter-sub.delivered)
+						if sub.consumeOpts.notifyOnReconnect {
+							sub.errs <- errConnected
 						}
+
 						sub.fetchNext <- &pullRequest{
 							Expires:   sub.consumeOpts.Expires,
-							Batch:     batchSize,
+							Batch:     sub.consumeOpts.MaxMessages,
 							MaxBytes:  sub.consumeOpts.MaxBytes,
 							Heartbeat: sub.consumeOpts.Heartbeat,
 						}
+						if sub.hbMonitor != nil {
+							sub.hbMonitor.Reset(2 * sub.consumeOpts.Heartbeat)
+						}
 						sub.resetPendingMsgs()
 					}
 					sub.Unlock()
@@ -325,9 +344,14 @@ func (p *pullConsumer) Consume(handler MessageHandler, opts ...PullConsumeOpt) (
 						MaxBytes:  sub.consumeOpts.MaxBytes,
 						Heartbeat: sub.consumeOpts.Heartbeat,
 					}
+					if sub.hbMonitor != nil {
+						sub.hbMonitor.Reset(2 * sub.consumeOpts.Heartbeat)
+					}
 					sub.resetPendingMsgs()
 				}
 				sub.Unlock()
+			case <-sub.done:
+				return
 			}
 		}
 	}()
@@ -364,10 +388,19 @@ func (s *pullSubscription) incrementDeliveredMsgs() {
 // the buffer to trigger a new pull request.
 // lock should be held before calling this method
 func (s *pullSubscription) checkPending() {
-	if s.pending.msgCount < s.consumeOpts.ThresholdMessages ||
-		(s.pending.byteCount < s.consumeOpts.ThresholdBytes && s.consumeOpts.MaxBytes != 0) &&
-			atomic.LoadUint32(&s.fetchInProgress) == 1 {
-		batchSize := s.consumeOpts.MaxMessages - s.pending.msgCount
+	if (s.pending.msgCount < s.consumeOpts.ThresholdMessages ||
+		(s.pending.byteCount < s.consumeOpts.ThresholdBytes && s.consumeOpts.MaxBytes != 0)) &&
+		s.fetchInProgress.Load() == 0 {
+
+		var batchSize, maxBytes int
+		if s.consumeOpts.MaxBytes == 0 {
+			// if using messages, calculate appropriate batch size
+			batchSize = s.consumeOpts.MaxMessages - s.pending.msgCount
+		} else {
+			// if using bytes, use the max value
+			batchSize = s.consumeOpts.MaxMessages
+			maxBytes = s.consumeOpts.MaxBytes - s.pending.byteCount
+		}
 		if s.consumeOpts.StopAfter > 0 {
 			batchSize = min(batchSize, s.consumeOpts.StopAfter-s.delivered-s.pending.msgCount)
 		}
@@ -375,7 +408,7 @@ func (s *pullSubscription) checkPending() {
 			s.fetchNext <- &pullRequest{
 				Expires:   s.consumeOpts.Expires,
 				Batch:     batchSize,
-				MaxBytes:  s.consumeOpts.MaxBytes - s.pending.byteCount,
+				MaxBytes:  maxBytes,
 				Heartbeat: s.consumeOpts.Heartbeat,
 			}
 
@@ -385,16 +418,13 @@ func (s *pullSubscription) checkPending() {
 	}
 }
 
-// Messages returns MessagesContext, allowing continuously iterating over messages on a stream.
+// Messages returns MessagesContext, allowing continuously iterating
+// over messages on a stream. Messages cannot be used concurrently
+// when using ordered consumer.
 //
-// Available options:
-// [PullMaxMessages] - sets maximum number of messages stored in a buffer, default is set to 100
-// [PullMaxBytes] - sets maximum number of bytes stored in a buffer
-// [PullExpiry] - sets a timeout for individual batch request, default is set to 30 seconds
-// [PullHeartbeat] - sets an idle heartbeat setting for a pull request, default is set to 5s
-// [WithMessagesErrOnMissingHeartbeat] - sets whether a missing heartbeat error should be reported when calling Next
+// See [Consumer.Messages] for more details.
 func (p *pullConsumer) Messages(opts ...PullMessagesOpt) (MessagesContext, error) {
-	consumeOpts, err := parseMessagesOpts(opts...)
+	consumeOpts, err := parseMessagesOpts(false, opts...)
 	if err != nil {
 		return nil, fmt.Errorf("%w: %s", ErrInvalidOption, err)
 	}
@@ -404,12 +434,7 @@ func (p *pullConsumer) Messages(opts ...PullMessagesOpt) (MessagesContext, error
 
 	msgs := make(chan *nats.Msg, consumeOpts.MaxMessages)
 
-	// for single consume, use empty string as id
-	// this is useful for ordered consumer, where only a single subscription is valid
-	var consumeID string
-	if len(p.subscriptions) > 0 {
-		consumeID = nuid.Next()
-	}
+	consumeID := nuid.Next()
 	sub := &pullSubscription{
 		id:          consumeID,
 		consumer:    p,
@@ -426,28 +451,40 @@ func (p *pullConsumer) Messages(opts ...PullMessagesOpt) (MessagesContext, error
 		p.Unlock()
 		return nil, err
 	}
+	sub.subscription.SetClosedHandler(func(sid string) func(string) {
+		return func(subject string) {
+			if sub.draining.Load() != 1 {
+				// if we're not draining, subscription can be closed as soon
+				// as closed handler is called
+				// otherwise, we need to wait until all messages are drained
+				// in Next
+				p.subs.Delete(sid)
+			}
+			close(msgs)
+		}
+	}(sub.id))
 
-	go func() {
-		<-sub.done
-		sub.cleanup()
-	}()
-	p.subscriptions[sub.id] = sub
+	p.subs.Store(sub.id, sub)
 	p.Unlock()
 
 	go sub.pullMessages(subject)
 
 	go func() {
 		for {
-			status, ok := <-sub.connStatusChanged
-			if !ok {
+			select {
+			case status, ok := <-sub.connStatusChanged:
+				if !ok {
+					return
+				}
+				if status == nats.CONNECTED {
+					sub.errs <- errConnected
+				}
+				if status == nats.RECONNECTING {
+					sub.errs <- errDisconnected
+				}
+			case <-sub.done:
 				return
 			}
-			if status == nats.CONNECTED {
-				sub.errs <- errConnected
-			}
-			if status == nats.RECONNECTING {
-				sub.errs <- errDisconnected
-			}
 		}
 	}()
 
@@ -459,13 +496,18 @@ var (
 	errDisconnected = errors.New("disconnected")
 )
 
+// Next retrieves next message on a stream. It will block until the next
+// message is available. If the context is canceled, Next will return
+// ErrMsgIteratorClosed error.
 func (s *pullSubscription) Next() (Msg, error) {
 	s.Lock()
 	defer s.Unlock()
-	if atomic.LoadUint32(&s.closed) == 1 {
+	drainMode := s.draining.Load() == 1
+	closed := s.closed.Load() == 1
+	if closed && !drainMode {
 		return nil, ErrMsgIteratorClosed
 	}
-	hbMonitor := s.scheduleHeartbeatCheck(s.consumeOpts.Heartbeat)
+	hbMonitor := s.scheduleHeartbeatCheck(2 * s.consumeOpts.Heartbeat)
 	defer func() {
 		if hbMonitor != nil {
 			hbMonitor.Stop()
@@ -481,9 +523,13 @@ func (s *pullSubscription) Next() (Msg, error) {
 	for {
 		s.checkPending()
 		select {
-		case <-s.done:
-			return nil, ErrMsgIteratorClosed
-		case msg := <-s.msgs:
+		case msg, ok := <-s.msgs:
+			if !ok {
+				// if msgs channel is closed, it means that subscription was either drained or stopped
+				s.consumer.subs.Delete(s.id)
+				s.draining.CompareAndSwap(1, 0)
+				return nil, ErrMsgIteratorClosed
+			}
 			if hbMonitor != nil {
 				hbMonitor.Reset(2 * s.consumeOpts.Heartbeat)
 			}
@@ -509,29 +555,22 @@ func (s *pullSubscription) Next() (Msg, error) {
 				if s.consumeOpts.ReportMissingHeartbeats {
 					return nil, err
 				}
+				if hbMonitor != nil {
+					hbMonitor.Reset(2 * s.consumeOpts.Heartbeat)
+				}
 			}
 			if errors.Is(err, errConnected) {
 				if !isConnected {
 					isConnected = true
-					// try fetching consumer info several times to make sure consumer is available after reconnect
-					for i := 0; i < 5; i++ {
-						ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
-						_, err := s.consumer.Info(ctx)
-						cancel()
-						if err == nil {
-							break
-						}
-						if err != nil {
-							if i == 4 {
-								s.Stop()
-								return nil, err
-							}
-						}
-						time.Sleep(5 * time.Second)
+
+					if s.consumeOpts.notifyOnReconnect {
+						return nil, errConnected
 					}
 					s.pending.msgCount = 0
 					s.pending.byteCount = 0
-					hbMonitor = s.scheduleHeartbeatCheck(s.consumeOpts.Heartbeat)
+					if hbMonitor != nil {
+						hbMonitor.Reset(2 * s.consumeOpts.Heartbeat)
+					}
 				}
 			}
 			if errors.Is(err, errDisconnected) {
@@ -546,12 +585,12 @@ func (s *pullSubscription) Next() (Msg, error) {
 
 func (s *pullSubscription) handleStatusMsg(msg *nats.Msg, msgErr error) error {
 	if !errors.Is(msgErr, nats.ErrTimeout) && !errors.Is(msgErr, ErrMaxBytesExceeded) {
-		if s.consumeOpts.ErrHandler != nil {
-			s.consumeOpts.ErrHandler(s, msgErr)
-		}
 		if errors.Is(msgErr, ErrConsumerDeleted) || errors.Is(msgErr, ErrBadRequest) {
 			return msgErr
 		}
+		if s.consumeOpts.ErrHandler != nil {
+			s.consumeOpts.ErrHandler(s, msgErr)
+		}
 		if errors.Is(msgErr, ErrConsumerLeadershipChanged) {
 			s.pending.msgCount = 0
 			s.pending.byteCount = 0
@@ -560,9 +599,7 @@ func (s *pullSubscription) handleStatusMsg(msg *nats.Msg, msgErr error) error {
 	}
 	msgsLeft, bytesLeft, err := parsePending(msg)
 	if err != nil {
-		if s.consumeOpts.ErrHandler != nil {
-			s.consumeOpts.ErrHandler(s, err)
-		}
+		return err
 	}
 	s.pending.msgCount -= msgsLeft
 	if s.pending.msgCount < 0 {
@@ -589,8 +626,11 @@ func (hb *hbMonitor) Reset(dur time.Duration) {
 	hb.Mutex.Unlock()
 }
 
+// Stop unsubscribes from the stream and cancels subscription. Calling
+// Next after calling Stop will return ErrMsgIteratorClosed error.
+// All messages that are already in the buffer are discarded.
 func (s *pullSubscription) Stop() {
-	if atomic.LoadUint32(&s.closed) == 1 {
+	if !s.closed.CompareAndSwap(0, 1) {
 		return
 	}
 	close(s.done)
@@ -601,54 +641,106 @@ func (s *pullSubscription) Stop() {
 			s.consumeOpts.stopAfterMsgsLeft <- s.consumeOpts.StopAfter - s.delivered
 		}
 	}
-	atomic.StoreUint32(&s.closed, 1)
+}
+
+// Drain unsubscribes from the stream and cancels subscription. All
+// messages that are already in the buffer will be available on
+// subsequent calls to Next. After the buffer is drained, Next will
+// return ErrMsgIteratorClosed error.
+func (s *pullSubscription) Drain() {
+	if !s.closed.CompareAndSwap(0, 1) {
+		return
+	}
+	s.draining.Store(1)
+	close(s.done)
+	if s.consumeOpts.stopAfterMsgsLeft != nil {
+		if s.delivered >= s.consumeOpts.StopAfter {
+			close(s.consumeOpts.stopAfterMsgsLeft)
+		} else {
+			s.consumeOpts.stopAfterMsgsLeft <- s.consumeOpts.StopAfter - s.delivered
+		}
+	}
+}
+
+// Closed returns a channel that is closed when consuming is
+// fully stopped/drained. When the channel is closed, no more messages
+// will be received and processing is complete.
+func (s *pullSubscription) Closed() <-chan struct{} {
+	s.Lock()
+	defer s.Unlock()
+	closedCh := s.closedCh
+	if closedCh == nil {
+		closedCh = make(chan struct{})
+		s.closedCh = closedCh
+	}
+	if !s.subscription.IsValid() {
+		close(s.closedCh)
+		s.closedCh = nil
+	}
+	return closedCh
 }
 
 // Fetch sends a single request to retrieve given number of messages.
 // It will wait up to provided expiry time if not all messages are available.
 func (p *pullConsumer) Fetch(batch int, opts ...FetchOpt) (MessageBatch, error) {
 	req := &pullRequest{
-		Batch:   batch,
-		Expires: DefaultExpires,
+		Batch:     batch,
+		Expires:   DefaultExpires,
+		Heartbeat: unset,
 	}
 	for _, opt := range opts {
 		if err := opt(req); err != nil {
 			return nil, err
 		}
 	}
-	// for longer pulls, set heartbeat value
-	if req.Expires >= 10*time.Second {
-		req.Heartbeat = 5 * time.Second
+	// if heartbeat was not explicitly set, set it to 5 seconds for longer pulls
+	// and disable it for shorter pulls
+	if req.Heartbeat == unset {
+		if req.Expires >= 10*time.Second {
+			req.Heartbeat = 5 * time.Second
+		} else {
+			req.Heartbeat = 0
+		}
+	}
+	if req.Expires < 2*req.Heartbeat {
+		return nil, fmt.Errorf("%w: expiry time should be at least 2 times the heartbeat", ErrInvalidOption)
 	}
 
 	return p.fetch(req)
 }
 
 // FetchBytes is used to retrieve up to a provided bytes from the stream.
-// This method will always send a single request and wait until provided number of bytes is
-// exceeded or request times out.
 func (p *pullConsumer) FetchBytes(maxBytes int, opts ...FetchOpt) (MessageBatch, error) {
 	req := &pullRequest{
-		Batch:    1000000,
-		MaxBytes: maxBytes,
-		Expires:  DefaultExpires,
+		Batch:     1000000,
+		MaxBytes:  maxBytes,
+		Expires:   DefaultExpires,
+		Heartbeat: unset,
 	}
 	for _, opt := range opts {
 		if err := opt(req); err != nil {
 			return nil, err
 		}
 	}
-	// for longer pulls, set heartbeat value
-	if req.Expires >= 10*time.Second {
-		req.Heartbeat = 5 * time.Second
+	// if heartbeat was not explicitly set, set it to 5 seconds for longer pulls
+	// and disable it for shorter pulls
+	if req.Heartbeat == unset {
+		if req.Expires >= 10*time.Second {
+			req.Heartbeat = 5 * time.Second
+		} else {
+			req.Heartbeat = 0
+		}
+	}
+	if req.Expires < 2*req.Heartbeat {
+		return nil, fmt.Errorf("%w: expiry time should be at least 2 times the heartbeat", ErrInvalidOption)
 	}
 
 	return p.fetch(req)
 }
 
 // FetchNoWait sends a single request to retrieve given number of messages.
-// If there are any messages available at the time of sending request,
-// FetchNoWait will return immediately.
+// FetchNoWait will only return messages that are available at the time of the
+// request. It will not wait for more messages to arrive.
 func (p *pullConsumer) FetchNoWait(batch int) (MessageBatch, error) {
 	req := &pullRequest{
 		Batch:  batch,
@@ -687,14 +779,9 @@ func (p *pullConsumer) fetch(req *pullRequest) (MessageBatch, error) {
 		defer sub.subscription.Unsubscribe()
 		defer close(res.msgs)
 		for {
-			if receivedMsgs == req.Batch || (req.MaxBytes != 0 && receivedBytes == req.MaxBytes) {
-				p.Lock()
-				res.done = true
-				p.Unlock()
-				return
-			}
 			select {
 			case msg := <-msgs:
+				res.Lock()
 				if hbTimer != nil {
 					hbTimer.Reset(2 * req.Heartbeat)
 				}
@@ -705,9 +792,11 @@ func (p *pullConsumer) fetch(req *pullRequest) (MessageBatch, error) {
 						res.err = err
 					}
 					res.done = true
+					res.Unlock()
 					return
 				}
 				if !userMsg {
+					res.Unlock()
 					continue
 				}
 				res.msgs <- p.jetStream.toJSMsg(msg)
@@ -720,8 +809,22 @@ func (p *pullConsumer) fetch(req *pullRequest) (MessageBatch, error) {
 				if req.MaxBytes != 0 {
 					receivedBytes += msg.Size()
 				}
-			case <-time.After(req.Expires + 1*time.Second):
+				if receivedMsgs == req.Batch || (req.MaxBytes != 0 && receivedBytes >= req.MaxBytes) {
+					res.done = true
+					res.Unlock()
+					return
+				}
+				res.Unlock()
+			case err := <-sub.errs:
+				res.Lock()
+				res.err = err
 				res.done = true
+				res.Unlock()
+				return
+			case <-time.After(req.Expires + 1*time.Second):
+				res.Lock()
+				res.done = true
+				res.Unlock()
 				return
 			}
 		}
@@ -730,13 +833,26 @@ func (p *pullConsumer) fetch(req *pullRequest) (MessageBatch, error) {
 }
 
 func (fr *fetchResult) Messages() <-chan Msg {
+	fr.Lock()
+	defer fr.Unlock()
 	return fr.msgs
 }
 
 func (fr *fetchResult) Error() error {
+	fr.Lock()
+	defer fr.Unlock()
 	return fr.err
 }
 
+func (fr *fetchResult) closed() bool {
+	fr.Lock()
+	defer fr.Unlock()
+	return fr.done
+}
+
+// Next is used to retrieve the next message from the stream. This
+// method will block until the message is retrieved or timeout is
+// reached.
 func (p *pullConsumer) Next(opts ...FetchOpt) (Msg, error) {
 	res, err := p.Fetch(1, opts...)
 	if err != nil {
@@ -756,7 +872,7 @@ func (s *pullSubscription) pullMessages(subject string) {
 	for {
 		select {
 		case req := <-s.fetchNext:
-			atomic.StoreUint32(&s.fetchInProgress, 1)
+			s.fetchInProgress.Store(1)
 
 			if err := s.pull(req, subject); err != nil {
 				if errors.Is(err, ErrMsgIteratorClosed) {
@@ -765,7 +881,7 @@ func (s *pullSubscription) pullMessages(subject string) {
 				}
 				s.errs <- err
 			}
-			atomic.StoreUint32(&s.fetchInProgress, 0)
+			s.fetchInProgress.Store(0)
 		case <-s.done:
 			s.cleanup()
 			return
@@ -785,18 +901,24 @@ func (s *pullSubscription) scheduleHeartbeatCheck(dur time.Duration) *hbMonitor
 }
 
 func (s *pullSubscription) cleanup() {
-	s.consumer.Lock()
-	defer s.consumer.Unlock()
-	if s.subscription == nil {
+	// For now this function does not need to hold the lock.
+	// Holding the lock here might cause a deadlock if Next()
+	// is already holding the lock and waiting.
+	// The fields that are read (subscription, hbMonitor)
+	// are read only (Only written on creation of pullSubscription).
+	if s.subscription == nil || !s.subscription.IsValid() {
 		return
 	}
 	if s.hbMonitor != nil {
 		s.hbMonitor.Stop()
 	}
-	s.subscription.Unsubscribe()
-	close(s.connStatusChanged)
-	s.subscription = nil
-	delete(s.consumer.subscriptions, s.id)
+	drainMode := s.draining.Load() == 1
+	if drainMode {
+		s.subscription.Drain()
+	} else {
+		s.subscription.Unsubscribe()
+	}
+	s.closed.Store(1)
 }
 
 // pull sends a pull request to the server and waits for messages using a subscription from [pullSubscription].
@@ -804,7 +926,7 @@ func (s *pullSubscription) cleanup() {
 func (s *pullSubscription) pull(req *pullRequest, subject string) error {
 	s.consumer.Lock()
 	defer s.consumer.Unlock()
-	if atomic.LoadUint32(&s.closed) == 1 {
+	if s.closed.Load() == 1 {
 		return ErrMsgIteratorClosed
 	}
 	if req.Batch < 1 {
@@ -822,7 +944,7 @@ func (s *pullSubscription) pull(req *pullRequest, subject string) error {
 	return nil
 }
 
-func parseConsumeOpts(opts ...PullConsumeOpt) (*consumeOpts, error) {
+func parseConsumeOpts(ordered bool, opts ...PullConsumeOpt) (*consumeOpts, error) {
 	consumeOpts := &consumeOpts{
 		MaxMessages:             unset,
 		MaxBytes:                unset,
@@ -836,13 +958,13 @@ func parseConsumeOpts(opts ...PullConsumeOpt) (*consumeOpts, error) {
 			return nil, err
 		}
 	}
-	if err := consumeOpts.setDefaults(); err != nil {
+	if err := consumeOpts.setDefaults(ordered); err != nil {
 		return nil, err
 	}
 	return consumeOpts, nil
 }
 
-func parseMessagesOpts(opts ...PullMessagesOpt) (*consumeOpts, error) {
+func parseMessagesOpts(ordered bool, opts ...PullMessagesOpt) (*consumeOpts, error) {
 	consumeOpts := &consumeOpts{
 		MaxMessages:             unset,
 		MaxBytes:                unset,
@@ -856,15 +978,15 @@ func parseMessagesOpts(opts ...PullMessagesOpt) (*consumeOpts, error) {
 			return nil, err
 		}
 	}
-	if err := consumeOpts.setDefaults(); err != nil {
+	if err := consumeOpts.setDefaults(ordered); err != nil {
 		return nil, err
 	}
 	return consumeOpts, nil
 }
 
-func (consumeOpts *consumeOpts) setDefaults() error {
+func (consumeOpts *consumeOpts) setDefaults(ordered bool) error {
 	if consumeOpts.MaxBytes != unset && consumeOpts.MaxMessages != unset {
-		return fmt.Errorf("only one of MaxMessages and MaxBytes can be specified")
+		return errors.New("only one of MaxMessages and MaxBytes can be specified")
 	}
 	if consumeOpts.MaxBytes != unset {
 		// when max_bytes is used, set batch size to a very large number
@@ -887,13 +1009,20 @@ func (consumeOpts *consumeOpts) setDefaults() error {
 		consumeOpts.ThresholdBytes = int(math.Ceil(float64(consumeOpts.MaxBytes) / 2))
 	}
 	if consumeOpts.Heartbeat == unset {
-		consumeOpts.Heartbeat = consumeOpts.Expires / 2
-		if consumeOpts.Heartbeat > 30*time.Second {
-			consumeOpts.Heartbeat = 30 * time.Second
+		if ordered {
+			consumeOpts.Heartbeat = 5 * time.Second
+			if consumeOpts.Expires < 10*time.Second {
+				consumeOpts.Heartbeat = consumeOpts.Expires / 2
+			}
+		} else {
+			consumeOpts.Heartbeat = consumeOpts.Expires / 2
+			if consumeOpts.Heartbeat > 30*time.Second {
+				consumeOpts.Heartbeat = 30 * time.Second
+			}
 		}
 	}
 	if consumeOpts.Heartbeat > consumeOpts.Expires/2 {
-		return fmt.Errorf("the value of Heartbeat must be less than 50%% of expiry")
+		return errors.New("the value of Heartbeat must be less than 50%% of expiry")
 	}
 	return nil
 }
