@@ -580,131 +580,123 @@ func TestOrderedConsumerConsume(t *testing.T) {
 	})
 
 	t.Run("wait for closed after drain", func(t *testing.T) {
-		for i := 0; i < 10; i++ {
-			t.Run(fmt.Sprintf("run %d", i), func(t *testing.T) {
-				srv := RunBasicJetStreamServer()
-				defer shutdownJSServerAndRemoveStorage(t, srv)
-				nc, err := nats.Connect(srv.ClientURL())
-				if err != nil {
-					t.Fatalf("Unexpected error: %v", err)
-				}
+		srv := RunBasicJetStreamServer()
+		defer shutdownJSServerAndRemoveStorage(t, srv)
+		nc, err := nats.Connect(srv.ClientURL())
+		if err != nil {
+			t.Fatalf("Unexpected error: %v", err)
+		}
 
-				js, err := jetstream.New(nc)
-				if err != nil {
-					t.Fatalf("Unexpected error: %v", err)
-				}
-				defer nc.Close()
-				ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
-				defer cancel()
-				s, err := js.CreateStream(ctx, jetstream.StreamConfig{Name: "foo", Subjects: []string{"FOO.*"}})
-				if err != nil {
-					t.Fatalf("Unexpected error: %v", err)
-				}
-				c, err := s.OrderedConsumer(ctx, jetstream.OrderedConsumerConfig{})
-				if err != nil {
-					t.Fatalf("Unexpected error: %v", err)
-				}
-				msgs := make([]jetstream.Msg, 0)
-				lock := sync.Mutex{}
-				publishTestMsgs(t, js)
-				cc, err := c.Consume(func(msg jetstream.Msg) {
-					time.Sleep(50 * time.Millisecond)
-					msg.Ack()
-					lock.Lock()
-					msgs = append(msgs, msg)
-					lock.Unlock()
-				})
-				if err != nil {
-					t.Fatalf("Unexpected error: %v", err)
-				}
-				closed := cc.Closed()
-				time.Sleep(100 * time.Millisecond)
-				if err := s.DeleteConsumer(context.Background(), c.CachedInfo().Name); err != nil {
-					t.Fatalf("Unexpected error: %v", err)
-				}
-				publishTestMsgs(t, js)
+		js, err := jetstream.New(nc)
+		if err != nil {
+			t.Fatalf("Unexpected error: %v", err)
+		}
+		defer nc.Close()
+		ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
+		defer cancel()
+		s, err := js.CreateStream(ctx, jetstream.StreamConfig{Name: "foo", Subjects: []string{"FOO.*"}})
+		if err != nil {
+			t.Fatalf("Unexpected error: %v", err)
+		}
+		c, err := s.OrderedConsumer(ctx, jetstream.OrderedConsumerConfig{})
+		if err != nil {
+			t.Fatalf("Unexpected error: %v", err)
+		}
+		msgs := make([]jetstream.Msg, 0)
+		lock := sync.Mutex{}
+		publishTestMsgs(t, js)
+		cc, err := c.Consume(func(msg jetstream.Msg) {
+			time.Sleep(50 * time.Millisecond)
+			msg.Ack()
+			lock.Lock()
+			msgs = append(msgs, msg)
+			lock.Unlock()
+		})
+		if err != nil {
+			t.Fatalf("Unexpected error: %v", err)
+		}
+		closed := cc.Closed()
+		time.Sleep(100 * time.Millisecond)
+		if err := s.DeleteConsumer(context.Background(), c.CachedInfo().Name); err != nil {
+			t.Fatalf("Unexpected error: %v", err)
+		}
+		publishTestMsgs(t, js)
 
-				// wait for the consumer to be recreated before calling drain
-				for i := 0; i < 5; i++ {
-					_, err = c.Info(ctx)
-					if err != nil {
-						if errors.Is(err, jetstream.ErrConsumerNotFound) {
-							time.Sleep(100 * time.Millisecond)
-							continue
-						}
-						t.Fatalf("Unexpected error: %v", err)
-					}
-					break
+		// wait for the consumer to be recreated before calling drain
+		for i := 0; i < 5; i++ {
+			_, err = c.Info(ctx)
+			if err != nil {
+				if errors.Is(err, jetstream.ErrConsumerNotFound) {
+					time.Sleep(100 * time.Millisecond)
+					continue
 				}
+				t.Fatalf("Unexpected error: %v", err)
+			}
+			break
+		}
 
-				cc.Drain()
+		cc.Drain()
 
-				select {
-				case <-closed:
-				case <-time.After(5 * time.Second):
-					t.Fatalf("Timeout waiting for consume to be closed")
-				}
+		select {
+		case <-closed:
+		case <-time.After(5 * time.Second):
+			t.Fatalf("Timeout waiting for consume to be closed")
+		}
 
-				if len(msgs) != 2*len(testMsgs) {
-					t.Fatalf("Unexpected received message count after consume closed; want %d; got %d", 2*len(testMsgs), len(msgs))
-				}
-			})
+		if len(msgs) != 2*len(testMsgs) {
+			t.Fatalf("Unexpected received message count after consume closed; want %d; got %d", 2*len(testMsgs), len(msgs))
 		}
 	})
 
 	t.Run("wait for closed on already closed consume", func(t *testing.T) {
-		for i := 0; i < 10; i++ {
-			t.Run(fmt.Sprintf("run %d", i), func(t *testing.T) {
-				srv := RunBasicJetStreamServer()
-				defer shutdownJSServerAndRemoveStorage(t, srv)
-				nc, err := nats.Connect(srv.ClientURL())
-				if err != nil {
-					t.Fatalf("Unexpected error: %v", err)
-				}
+		srv := RunBasicJetStreamServer()
+		defer shutdownJSServerAndRemoveStorage(t, srv)
+		nc, err := nats.Connect(srv.ClientURL())
+		if err != nil {
+			t.Fatalf("Unexpected error: %v", err)
+		}
 
-				js, err := jetstream.New(nc)
-				if err != nil {
-					t.Fatalf("Unexpected error: %v", err)
-				}
-				defer nc.Close()
-				ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
-				defer cancel()
-				s, err := js.CreateStream(ctx, jetstream.StreamConfig{Name: "foo", Subjects: []string{"FOO.*"}})
-				if err != nil {
-					t.Fatalf("Unexpected error: %v", err)
-				}
-				c, err := s.OrderedConsumer(ctx, jetstream.OrderedConsumerConfig{})
-				if err != nil {
-					t.Fatalf("Unexpected error: %v", err)
-				}
-				msgs := make([]jetstream.Msg, 0)
-				lock := sync.Mutex{}
-				publishTestMsgs(t, js)
-				cc, err := c.Consume(func(msg jetstream.Msg) {
-					time.Sleep(50 * time.Millisecond)
-					msg.Ack()
-					lock.Lock()
-					msgs = append(msgs, msg)
-					lock.Unlock()
-				})
-				if err != nil {
-					t.Fatalf("Unexpected error: %v", err)
-				}
-				time.Sleep(100 * time.Millisecond)
-				if err := s.DeleteConsumer(context.Background(), c.CachedInfo().Name); err != nil {
-					t.Fatalf("Unexpected error: %v", err)
-				}
+		js, err := jetstream.New(nc)
+		if err != nil {
+			t.Fatalf("Unexpected error: %v", err)
+		}
+		defer nc.Close()
+		ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
+		defer cancel()
+		s, err := js.CreateStream(ctx, jetstream.StreamConfig{Name: "foo", Subjects: []string{"FOO.*"}})
+		if err != nil {
+			t.Fatalf("Unexpected error: %v", err)
+		}
+		c, err := s.OrderedConsumer(ctx, jetstream.OrderedConsumerConfig{})
+		if err != nil {
+			t.Fatalf("Unexpected error: %v", err)
+		}
+		msgs := make([]jetstream.Msg, 0)
+		lock := sync.Mutex{}
+		publishTestMsgs(t, js)
+		cc, err := c.Consume(func(msg jetstream.Msg) {
+			time.Sleep(50 * time.Millisecond)
+			msg.Ack()
+			lock.Lock()
+			msgs = append(msgs, msg)
+			lock.Unlock()
+		})
+		if err != nil {
+			t.Fatalf("Unexpected error: %v", err)
+		}
+		time.Sleep(100 * time.Millisecond)
+		if err := s.DeleteConsumer(context.Background(), c.CachedInfo().Name); err != nil {
+			t.Fatalf("Unexpected error: %v", err)
+		}
 
-				cc.Stop()
+		cc.Stop()
 
-				time.Sleep(100 * time.Millisecond)
+		time.Sleep(100 * time.Millisecond)
 
-				select {
-				case <-cc.Closed():
-				case <-time.After(5 * time.Second):
-					t.Fatalf("Timeout waiting for consume to be closed")
-				}
-			})
+		select {
+		case <-cc.Closed():
+		case <-time.After(5 * time.Second):
+			t.Fatalf("Timeout waiting for consume to be closed")
 		}
 	})
 }
@@ -1571,21 +1563,19 @@ func TestOrderedConsumerNext(t *testing.T) {
 		}
 
 		publishTestMsgs(t, js)
-		msg, err := c.Next()
+		_, err = c.Next()
 		if err != nil {
 			t.Fatalf("Unexpected error: %s", err)
 		}
-		msg.Ack()
 
 		name := c.CachedInfo().Name
 		if err := s.DeleteConsumer(ctx, name); err != nil {
 			t.Fatal(err)
 		}
-		msg, err = c.Next()
+		_, err = c.Next()
 		if err != nil {
 			t.Fatalf("Unexpected error: %s", err)
 		}
-		msg.Ack()
 	})
 
 	t.Run("consumer used as consume", func(t *testing.T) {
@@ -1623,6 +1613,70 @@ func TestOrderedConsumerNext(t *testing.T) {
 			t.Fatalf("Expected error: %s; got: %s", jetstream.ErrOrderConsumerUsedAsConsume, err)
 		}
 	})
+
+	t.Run("preserve sequence after fetch error", func(t *testing.T) {
+		srv := RunBasicJetStreamServer()
+		defer shutdownJSServerAndRemoveStorage(t, srv)
+		nc, err := nats.Connect(srv.ClientURL())
+		if err != nil {
+			t.Fatalf("Unexpected error: %v", err)
+		}
+
+		js, err := jetstream.New(nc)
+		if err != nil {
+			t.Fatalf("Unexpected error: %v", err)
+		}
+		defer nc.Close()
+
+		ctx, cancel := context.WithTimeout(context.Background(), 20*time.Second)
+		defer cancel()
+		s, err := js.CreateStream(ctx, jetstream.StreamConfig{Name: "foo", Subjects: []string{"FOO.*"}})
+		if err != nil {
+			t.Fatalf("Unexpected error: %v", err)
+		}
+		c, err := s.OrderedConsumer(ctx, jetstream.OrderedConsumerConfig{})
+		if err != nil {
+			t.Fatalf("Unexpected error: %v", err)
+		}
+
+		if _, err := js.Publish(ctx, "FOO.A", []byte("msg")); err != nil {
+			t.Fatalf("Unexpected error during publish: %s", err)
+		}
+		msg, err := c.Next()
+		if err != nil {
+			t.Fatalf("Unexpected error: %s", err)
+		}
+		meta, err := msg.Metadata()
+		if err != nil {
+			t.Fatalf("Unexpected error: %s", err)
+		}
+		if meta.Sequence.Stream != 1 {
+			t.Fatalf("Expected sequence: %d; got: %d", 1, meta.Sequence.Stream)
+		}
+
+		// get next message, it should time out (no more messages on stream)
+		_, err = c.Next(jetstream.FetchMaxWait(100 * time.Millisecond))
+		if !errors.Is(err, nats.ErrTimeout) {
+			t.Fatalf("Expected error: %s; got: %s", nats.ErrTimeout, err)
+		}
+
+		if _, err := js.Publish(ctx, "FOO.A", []byte("msg")); err != nil {
+			t.Fatalf("Unexpected error during publish: %s", err)
+		}
+
+		// get next message, it should have stream sequence 2
+		msg, err = c.Next()
+		if err != nil {
+			t.Fatalf("Unexpected error: %s", err)
+		}
+		meta, err = msg.Metadata()
+		if err != nil {
+			t.Fatalf("Unexpected error: %s", err)
+		}
+		if meta.Sequence.Stream != 2 {
+			t.Fatalf("Expected sequence: %d; got: %d", 2, meta.Sequence.Stream)
+		}
+	})
 }
 
 func TestOrderedConsumerFetchNoWait(t *testing.T) {
@@ -1989,6 +2043,56 @@ func TestOrderedConsumerConfig(t *testing.T) {
 				HeadersOnly:       true,
 			},
 		},
+		{
+			name: "both start seq and time set, deliver policy start seq",
+			config: jetstream.OrderedConsumerConfig{
+				FilterSubjects:    []string{"foo.a", "foo.b"},
+				DeliverPolicy:     jetstream.DeliverByStartSequencePolicy,
+				OptStartSeq:       10,
+				OptStartTime:      &time.Time{},
+				ReplayPolicy:      jetstream.ReplayOriginalPolicy,
+				InactiveThreshold: 10 * time.Second,
+				HeadersOnly:       true,
+			},
+			expected: jetstream.ConsumerConfig{
+				FilterSubjects:    []string{"foo.a", "foo.b"},
+				OptStartSeq:       10,
+				OptStartTime:      nil,
+				DeliverPolicy:     jetstream.DeliverByStartSequencePolicy,
+				AckPolicy:         jetstream.AckNonePolicy,
+				MaxDeliver:        -1,
+				MaxWaiting:        512,
+				InactiveThreshold: 10 * time.Second,
+				Replicas:          1,
+				MemoryStorage:     true,
+				HeadersOnly:       true,
+			},
+		},
+		{
+			name: "both start seq and time set, deliver policy start time",
+			config: jetstream.OrderedConsumerConfig{
+				FilterSubjects:    []string{"foo.a", "foo.b"},
+				DeliverPolicy:     jetstream.DeliverByStartTimePolicy,
+				OptStartSeq:       10,
+				OptStartTime:      &time.Time{},
+				ReplayPolicy:      jetstream.ReplayOriginalPolicy,
+				InactiveThreshold: 10 * time.Second,
+				HeadersOnly:       true,
+			},
+			expected: jetstream.ConsumerConfig{
+				FilterSubjects:    []string{"foo.a", "foo.b"},
+				OptStartSeq:       0,
+				OptStartTime:      &time.Time{},
+				DeliverPolicy:     jetstream.DeliverByStartTimePolicy,
+				AckPolicy:         jetstream.AckNonePolicy,
+				MaxDeliver:        -1,
+				MaxWaiting:        512,
+				InactiveThreshold: 10 * time.Second,
+				Replicas:          1,
+				MemoryStorage:     true,
+				HeadersOnly:       true,
+			},
+		},
 	}
 
 	for _, test := range tests {
