@@ -16,8 +16,10 @@ package test
 import (
 	"context"
 	"errors"
+	"fmt"
 	"os"
 	"reflect"
+	"sync"
 	"testing"
 	"time"
 
@@ -1075,8 +1077,59 @@ func TestPublishMsgAsync(t *testing.T) {
 						Subject: "ABC",
 					},
 					withAckError: func(t *testing.T, err error) {
-						if !errors.Is(err, nats.ErrNoResponders) {
-							t.Fatalf("Expected error: %v; got: %v", nats.ErrNoResponders, err)
+						if !errors.Is(err, jetstream.ErrNoStreamResponse) {
+							t.Fatalf("Expected error: %v; got: %v", jetstream.ErrNoStreamResponse, err)
+						}
+					},
+				},
+			},
+		},
+		{
+			name: "invalid retry number set",
+			msgs: []publishConfig{
+				{
+					msg: &nats.Msg{
+						Data:    []byte("msg 1"),
+						Subject: "FOO.1",
+					},
+					opts: []jetstream.PublishOpt{jetstream.WithRetryAttempts(-1)},
+					withPublishError: func(t *testing.T, err error) {
+						if !errors.Is(err, jetstream.ErrInvalidOption) {
+							t.Fatalf("Expected error: %v; got: %v", jetstream.ErrInvalidOption, err)
+						}
+					},
+				},
+			},
+		},
+		{
+			name: "invalid retry wait set",
+			msgs: []publishConfig{
+				{
+					msg: &nats.Msg{
+						Data:    []byte("msg 1"),
+						Subject: "FOO.1",
+					},
+					opts: []jetstream.PublishOpt{jetstream.WithRetryWait(-1)},
+					withPublishError: func(t *testing.T, err error) {
+						if !errors.Is(err, jetstream.ErrInvalidOption) {
+							t.Fatalf("Expected error: %v; got: %v", jetstream.ErrInvalidOption, err)
+						}
+					},
+				},
+			},
+		},
+		{
+			name: "invalid stall wait set",
+			msgs: []publishConfig{
+				{
+					msg: &nats.Msg{
+						Data:    []byte("msg 1"),
+						Subject: "FOO.1",
+					},
+					opts: []jetstream.PublishOpt{jetstream.WithStallWait(-1)},
+					withPublishError: func(t *testing.T, err error) {
+						if !errors.Is(err, jetstream.ErrInvalidOption) {
+							t.Fatalf("Expected error: %v; got: %v", jetstream.ErrInvalidOption, err)
 						}
 					},
 				},
@@ -1218,18 +1271,23 @@ func TestPublishMsgAsyncWithPendingMsgs(t *testing.T) {
 		ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
 		defer cancel()
 
-		_, err = js.CreateStream(ctx, jetstream.StreamConfig{Name: "foo", Subjects: []string{"FOO.*"}})
+		_, err = js.CreateStream(ctx, jetstream.StreamConfig{
+			Name:     "foo",
+			Subjects: []string{"FOO.*"},
+			// disable stream acks
+			NoAck: true,
+		})
 		if err != nil {
 			t.Fatalf("Unexpected error: %v", err)
 		}
 
 		for i := 0; i < 5; i++ {
-			_, err = js.PublishAsync("FOO.1", []byte("msg"), jetstream.WithStallWait(1*time.Nanosecond))
+			_, err = js.PublishAsync("FOO.1", []byte("msg"), jetstream.WithStallWait(10*time.Millisecond))
 			if err != nil {
 				t.Fatalf("Unexpected error: %v", err)
 			}
 		}
-		if _, err = js.PublishAsync("FOO.1", []byte("msg"), jetstream.WithStallWait(1*time.Nanosecond)); err == nil || !errors.Is(err, jetstream.ErrTooManyStalledMsgs) {
+		if _, err = js.PublishAsync("FOO.1", []byte("msg"), jetstream.WithStallWait(10*time.Millisecond)); err == nil || !errors.Is(err, jetstream.ErrTooManyStalledMsgs) {
 			t.Fatalf("Expected error: %v; got: %v", jetstream.ErrTooManyStalledMsgs, err)
 		}
 	})
@@ -1274,7 +1332,6 @@ func TestPublishMsgAsyncWithPendingMsgs(t *testing.T) {
 
 func TestPublishAsyncResetPendingOnReconnect(t *testing.T) {
 	s := RunBasicJetStreamServer()
-	defer shutdownJSServerAndRemoveStorage(t, s)
 
 	nc, err := nats.Connect(s.ClientURL())
 	if err != nil {
@@ -1296,6 +1353,7 @@ func TestPublishAsyncResetPendingOnReconnect(t *testing.T) {
 	errs := make(chan error, 1)
 	done := make(chan struct{}, 1)
 	acks := make(chan jetstream.PubAckFuture, 100)
+	wg := sync.WaitGroup{}
 	go func() {
 		for i := 0; i < 100; i++ {
 			if ack, err := js.PublishAsync("FOO.A", []byte("hello")); err != nil {
@@ -1304,6 +1362,7 @@ func TestPublishAsyncResetPendingOnReconnect(t *testing.T) {
 			} else {
 				acks <- ack
 			}
+			wg.Add(1)
 		}
 		close(acks)
 		done <- struct{}{}
@@ -1315,23 +1374,181 @@ func TestPublishAsyncResetPendingOnReconnect(t *testing.T) {
 	case <-time.After(5 * time.Second):
 		t.Fatalf("Did not receive completion signal")
 	}
-	s.Shutdown()
-	time.Sleep(100 * time.Millisecond)
-	if pending := js.PublishAsyncPending(); pending != 0 {
-		t.Fatalf("Expected no pending messages after server shutdown; got: %d", pending)
+	for ack := range acks {
+		go func(paf jetstream.PubAckFuture) {
+			select {
+			case <-paf.Ok():
+			case err := <-paf.Err():
+				if !errors.Is(err, nats.ErrDisconnected) && !errors.Is(err, nats.ErrNoResponders) {
+					errs <- fmt.Errorf("Expected error: %v or %v; got: %v", nats.ErrDisconnected, nats.ErrNoResponders, err)
+				}
+			case <-time.After(5 * time.Second):
+				errs <- errors.New("Did not receive completion signal")
+			}
+			wg.Done()
+		}(ack)
 	}
-	s = RunBasicJetStreamServer()
+	s = restartBasicJSServer(t, s)
 	defer shutdownJSServerAndRemoveStorage(t, s)
 
-	for ack := range acks {
-		select {
-		case <-ack.Ok():
-		case err := <-ack.Err():
-			if !errors.Is(err, nats.ErrDisconnected) && !errors.Is(err, nats.ErrNoResponders) {
-				t.Fatalf("Expected error: %v or %v; got: %v", nats.ErrDisconnected, nats.ErrNoResponders, err)
-			}
-		case <-time.After(5 * time.Second):
-			t.Fatalf("Did not receive completion signal")
+	wg.Wait()
+	select {
+	case err := <-errs:
+		t.Fatalf("Unexpected error: %v", err)
+	default:
+	}
+}
+
+func TestPublishAsyncRetry(t *testing.T) {
+	tests := []struct {
+		name     string
+		pubOpts  []jetstream.PublishOpt
+		ackError error
+	}{
+		{
+			name: "retry until stream is ready",
+			pubOpts: []jetstream.PublishOpt{
+				jetstream.WithRetryAttempts(10),
+				jetstream.WithRetryWait(100 * time.Millisecond),
+			},
+		},
+		{
+			name: "fail after max retries",
+			pubOpts: []jetstream.PublishOpt{
+				jetstream.WithRetryAttempts(2),
+				jetstream.WithRetryWait(50 * time.Millisecond),
+			},
+			ackError: jetstream.ErrNoStreamResponse,
+		},
+		{
+			name: "retries disabled",
+			pubOpts: []jetstream.PublishOpt{
+				jetstream.WithRetryAttempts(0),
+			},
+			ackError: jetstream.ErrNoStreamResponse,
+		},
+	}
+
+	for _, test := range tests {
+		t.Run(test.name, func(t *testing.T) {
+			s := RunBasicJetStreamServer()
+			defer shutdownJSServerAndRemoveStorage(t, s)
+
+			nc, err := nats.Connect(s.ClientURL())
+			if err != nil {
+				t.Fatalf("Unexpected error: %v", err)
+			}
+
+			// set max pending to 1 so that we can test if retries don't cause stall
+			js, err := jetstream.New(nc, jetstream.WithPublishAsyncMaxPending(1))
+			if err != nil {
+				t.Fatalf("Unexpected error: %v", err)
+			}
+			defer nc.Close()
+			ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
+			defer cancel()
+
+			test.pubOpts = append(test.pubOpts, jetstream.WithStallWait(1*time.Nanosecond))
+			ack, err := js.PublishAsync("foo", []byte("hello"), test.pubOpts...)
+			if err != nil {
+				t.Fatalf("Unexpected error: %v", err)
+			}
+			publishComplete := js.PublishAsyncComplete()
+			errs := make(chan error, 1)
+			go func() {
+				// create stream with delay so that publish will receive no responders
+				time.Sleep(300 * time.Millisecond)
+				if _, err := js.CreateStream(ctx, jetstream.StreamConfig{Name: "TEST", Subjects: []string{"foo"}}); err != nil {
+					errs <- err
+				}
+			}()
+			select {
+			case <-ack.Ok():
+			case err := <-ack.Err():
+				if test.ackError != nil {
+					if !errors.Is(err, test.ackError) {
+						t.Fatalf("Expected error: %v; got: %v", test.ackError, err)
+					}
+				} else {
+					t.Fatalf("Unexpected ack error: %v", err)
+				}
+			case err := <-errs:
+				t.Fatalf("Error creating stream: %v", err)
+			case <-time.After(5 * time.Second):
+				t.Fatalf("Timeout waiting for ack")
+			}
+
+			select {
+			case <-publishComplete:
+			case <-time.After(5 * time.Second):
+				t.Fatalf("Did not receive completion signal")
+			}
+		})
+	}
+}
+
+func TestPublishAsyncRetryInErrHandler(t *testing.T) {
+	s := RunBasicJetStreamServer()
+	defer shutdownJSServerAndRemoveStorage(t, s)
+
+	nc, err := nats.Connect(s.ClientURL())
+	if err != nil {
+		t.Fatalf("Unexpected error: %v", err)
+	}
+
+	streamCreated := make(chan struct{})
+	errCB := func(js jetstream.JetStream, m *nats.Msg, e error) {
+		<-streamCreated
+		_, err := js.PublishMsgAsync(m)
+		if err != nil {
+			t.Fatalf("Unexpected error when republishing: %v", err)
+		}
+	}
+
+	js, err := jetstream.New(nc, jetstream.WithPublishAsyncErrHandler(errCB))
+	if err != nil {
+		t.Fatalf("Unexpected error: %v", err)
+	}
+	defer nc.Close()
+	ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
+	defer cancel()
+
+	errs := make(chan error, 1)
+	done := make(chan struct{}, 1)
+	go func() {
+		for i := 0; i < 10; i++ {
+			if _, err := js.PublishAsync("FOO.A", []byte("hello"), jetstream.WithRetryAttempts(0)); err != nil {
+				errs <- err
+				return
+			}
 		}
+		done <- struct{}{}
+	}()
+	select {
+	case <-done:
+	case err := <-errs:
+		t.Fatalf("Unexpected error during publish: %v", err)
+	case <-time.After(5 * time.Second):
+		t.Fatalf("Did not receive completion signal")
+	}
+	stream, err := js.CreateStream(ctx, jetstream.StreamConfig{Name: "foo", Subjects: []string{"FOO.*"}})
+	if err != nil {
+		t.Fatalf("Unexpected error: %v", err)
+	}
+
+	close(streamCreated)
+	select {
+	case <-js.PublishAsyncComplete():
+	case <-time.After(5 * time.Second):
+		t.Fatalf("Did not receive completion signal")
+	}
+
+	info, err := stream.Info(context.Background())
+	if err != nil {
+		t.Fatalf("Unexpected error: %v", err)
+	}
+
+	if info.State.Msgs != 10 {
+		t.Fatalf("Expected 10 messages in the stream; got: %d", info.State.Msgs)
 	}
 }
