@@ -21,6 +21,7 @@ import (
 	"fmt"
 	"math/rand"
 	"reflect"
+	"sync"
 	"testing"
 	"time"
 
@@ -180,7 +181,7 @@ func TestServiceBasics(t *testing.T) {
 
 	// Services should process 50 requests total
 	if requestsNum != 50 {
-		t.Fatalf("Expected a total fo 50 requests processed, got: %d", requestsNum)
+		t.Fatalf("Expected a total of 50 requests processed, got: %d", requestsNum)
 	}
 	// Reset stats for a service
 	svcs[0].Reset()
@@ -188,7 +189,6 @@ func TestServiceBasics(t *testing.T) {
 	if svcs[0].Stats().Endpoints[0].NumRequests != 0 {
 		t.Fatalf("Expected empty stats after reset; got: %+v", svcs[0].Stats())
 	}
-
 }
 
 func TestAddService(t *testing.T) {
@@ -504,7 +504,7 @@ func TestAddService(t *testing.T) {
 			}
 
 			if test.givenConfig.ErrorHandler != nil {
-				go nc.Opts.AsyncErrorCB(nc, &nats.Subscription{Subject: test.asyncErrorSubject}, fmt.Errorf("oops"))
+				go nc.Opts.AsyncErrorCB(nc, &nats.Subscription{Subject: test.asyncErrorSubject}, errors.New("oops"))
 				select {
 				case <-errService:
 				case <-time.After(1 * time.Second):
@@ -536,7 +536,7 @@ func TestAddService(t *testing.T) {
 				}
 			}
 			if test.natsErrorHandler != nil {
-				go nc.Opts.AsyncErrorCB(nc, &nats.Subscription{Subject: test.asyncErrorSubject}, fmt.Errorf("oops"))
+				go nc.Opts.AsyncErrorCB(nc, &nats.Subscription{Subject: test.asyncErrorSubject}, errors.New("oops"))
 				select {
 				case <-errService:
 					t.Fatalf("Expected to restore nats error handler")
@@ -634,7 +634,7 @@ func TestErrHandlerSubjectMatch(t *testing.T) {
 				}
 				defer svc.Stop()
 
-				go nc.Opts.AsyncErrorCB(nc, &nats.Subscription{Subject: test.errSubject}, fmt.Errorf("oops"))
+				go nc.Opts.AsyncErrorCB(nc, &nats.Subscription{Subject: test.errSubject}, errors.New("oops"))
 				if test.expectServiceErr {
 					select {
 					case <-errChan:
@@ -1016,6 +1016,64 @@ func TestContextHandler(t *testing.T) {
 	if resp.Header.Get(micro.ErrorCodeHeader) != "400" {
 		t.Fatalf("Expected error response after canceling context; got: %q", string(resp.Data))
 	}
+}
+
+func TestAddEndpoint_Concurrency(t *testing.T) {
+	s := RunServerOnPort(-1)
+	defer s.Shutdown()
+
+	nc, err := nats.Connect(s.ClientURL())
+	if err != nil {
+		t.Fatalf("Expected to connect to server, got %v", err)
+	}
+	defer nc.Close()
+
+	ctx := context.Background()
+
+	handler := func(ctx context.Context, req micro.Request) {
+		req.RespondJSON(map[string]any{"hello": "world"})
+	}
+	config := micro.Config{
+		Name:    "test_service",
+		Version: "0.1.0",
+	}
+
+	srv, err := micro.AddService(nc, config)
+	if err != nil {
+		t.Fatalf("Unexpected error: %v", err)
+	}
+	defer srv.Stop()
+
+	res := make(chan error, 10)
+	wg := sync.WaitGroup{}
+	wg.Add(10)
+
+	// now add a few endpoints concurrently
+	// and make sure they are added successfully
+	// and there is no race
+	for i := 0; i < 10; i++ {
+		go func(i int) {
+			wg.Wait()
+			res <- srv.AddEndpoint(fmt.Sprintf("test%d", i), micro.ContextHandler(ctx, handler))
+		}(i)
+		// after all goroutines are started, release the lock
+	}
+	wg.Add(-10)
+
+	for i := 0; i < 10; i++ {
+		select {
+		case err := <-res:
+			if err != nil {
+				t.Fatalf("Unexpected error: %s", err)
+			}
+		case <-time.After(1 * time.Second):
+			t.Fatalf("Timeout waiting for endpoint to be added")
+		}
+	}
+
+	if len(srv.Info().Endpoints) != 10 {
+		t.Fatalf("Expected 11 endpoints, got: %d", len(srv.Info().Endpoints))
+	}
 
 }
 
